#include <gtk--/main.h>
#include <gtk--/window.h>
#include <gtk--/button.h>
#include <gtk--/box.h>
#include <gtk--/drawingarea.h>
#include <gtk--/style.h>
#include <gtk--/label.h>
#include "errno.h"
#include <sys/stat.h>
#include "save-cwd.h"
#include "en_tree.h"
#include <unistd.h>
#include "savedir.h"
#include "config.h"
#include <fcntl.h>

#include <vector>
#include "dutree.h"

void destroy_handler()
{
	Gtk::Main::quit();
}

Rectangle::Rectangle(float left, float top, float width, float height)
{
	this->left=left;
	this->top=top;
	this->width=width;
	this->height=height;
}

bool Rectangle::contains(float x, float y)
{
	if (x>=left && y>=top)
	{
		if(x-left<=width && y-top<=height)
			return true;
	}
	return false;
}
TextRect::TextRect(TextRect const &arg):
	Rectangle(arg)
{
	my_en_tree=arg.my_en_tree;
}

void TextRect::setText(en_tree *arg_en_tree)
{
	my_en_tree=arg_en_tree;
}

void TextRect::setColor(Gdk_Color &curcolor)
{
	curcolor.set_red(my_en_tree->r);
	curcolor.set_green(my_en_tree->g);
	curcolor.set_blue(my_en_tree->b);
}

TreeView::TreeView(gint x_size = 0, gint y_size = 0)
{
	size(x_size, y_size);
	DirTree treeGen;
	//treeGen.max_depth=3;
	treeGen.opt_separate_dirs=false;
	//test_tree();
	my_tree=treeGen.get_en_tree(".");
	cur_tree=my_tree;
	firsttime=true;
	en_tree_rect(cur_tree, false, 0,0,300,300);
}

en_tree const *TreeView::get_en_tree()
{
	return cur_tree;
}

void TreeView::set_en_tree(en_tree *newcur)
{
	cur_tree=newcur;
	rectangles.clear();
	en_tree_rect(cur_tree, false, 0,0,300,300);
	srandom(1);
	draw_rect();
}

gint TreeView::expose_event_impl(GdkEventExpose*)
{
	srandom(1);
	firsttime=false;
	draw_rect();
}


/* Recursive en_tree drawing function */
void TreeView::en_tree_rect(en_tree *curent, bool vertical, gint x, gint y, 
	gint width, gint height)
{
	if (firsttime)
	{

	cout << curent->name << endl;
	cout << "x: " << x << " y: "<< y << " width: " << width << " height: " <<\
		height << endl;
	}

/* calculate the total size of children */
	uintmax_t childsize=0;
	for (en_tree *curchild=curent->child; curchild!=NULL; 
		curchild=curchild->next)
	{
		childsize+=curchild->size;
	}

/* calculate the size of the entry without children */
	uintmax_t selfsize=curent->size-childsize;
	
/* prepare the adjustable dimension and coordinate */	
	gint &dimension=(vertical)?width:height;
	gint &coordinate=(vertical)?x:y;
	gint olddimension=dimension;
	dimension=dimension*selfsize/curent->size;
	TextRect cur(x, y, width+1, height);
	cur.setText(curent);
	rectangles.push_back(cur);
/* draw the rectangle */	

	
	if (firsttime)
	{
	cout << "x: " << x << " y: "<< y << " width: " << width << " height: " <<\
		height << endl << endl;
	}

/* ajust the coordiate and dimension for drawing children */
	coordinate+=dimension;
	dimension=olddimension-dimension;

	/* prepare the adjustable (opposite) dimension and coordinate for the child  */
	gint &child_dimension=(vertical)?height:width;
	olddimension=child_dimension;
	gint &child_coordinate=(vertical)?y:x;
	float fcoord=child_coordinate;
	for (en_tree *curchild=curent->child; curchild!=NULL; 
		curchild=curchild->next)
	{
		float fdim;
		fdim=1.0f*child_dimension*curchild->size/childsize;
		child_dimension=static_cast<int>(fdim);
		en_tree_rect(curchild, vertical, x, y, width, height);
		fcoord+=fdim;
		child_coordinate=static_cast<int>(fcoord);
		if (firsttime)
		{
			cout << fcoord << "," << fdim<<endl;
		}
		child_dimension=olddimension;
	}
}

void TreeView::draw_rect()
{
	Gdk_GC context;
	context.create(get_window());
	Gdk_Color curcolor;
	Gdk_Colormap colormap(Gdk_Colormap::get_system());
	
	Gdk_Font text_font("10x20");

	for (size_t i=0; i<rectangles.size(); ++i)
	{
		TextRect &cur=rectangles[i];
		cur.setColor(curcolor);
		colormap.alloc(curcolor);
		context.set_foreground(curcolor);
		get_window().draw_rectangle(context, true, cur.left, cur.top, 
			cur.width+1, cur.height);
//get_window().draw_string(text_font, this->get_style()->get_black_gc(), x, 
	//	y+height/2, curent->name);


	}
}

void TreeView::test_tree()
{
	en_tree *child;
	my_tree=en_tree_new();
	en_tree_set(my_tree, 100, "Top");
	child=en_tree_new();
	en_tree_set(child, 25, "Yark");
	en_tree_add_child(my_tree, child);
	child=en_tree_new();
	en_tree_set(child, 12, "Park");
	en_tree_add_child(my_tree, child);
	en_tree *childchild=en_tree_new();
	en_tree_set(childchild, 6, "Park");
	en_tree_add_child(child, childchild);
	child=en_tree_new();
	en_tree_set(child, 13, "Sark");
	en_tree_add_child(my_tree, child); 
}

static void
ensure_space (String *s, unsigned int size)
{
  if (s->alloc < size)
    {
      s->text = static_cast<char *>(xrealloc (s->text, size + 1));
      s->alloc = size;
    }
}
/* Initialize string S1 to hold SIZE characters.  */

static void
str_init (String **s1, unsigned int size)
{
  String *s;

  s = (String *) xmalloc (sizeof (struct String));
  s->text = static_cast<char *>(xmalloc (size + 1));

  s->alloc = size;
  *s1 = s;
}

/* Restore the previous working directory or exit.
   If CWD is null, simply call `chdir ("..")'.  Otherwise,
   use CWD and free it.  CURR_DIR_NAME is the name of the current directory
   and is used solely in failure diagnostics.  */
static void str_concatc (String *s1, const char *cstr)
{
  unsigned l1 = s1->length;
  unsigned l2 = strlen (cstr);
  unsigned l = l1 + l2;

  ensure_space (s1, l);
  strcpy (s1->text + l1, cstr);
  s1->length = l;
}

/* Truncate the string S1 to have length LENGTH.  */

static void
str_trunc (String *s1, unsigned int length)
{
  if (s1->length > length)
    {
      s1->text[length] = 0;
      s1->length = length;
    }
}

static void
pop_dir (struct saved_cwd *cwd, const char *curr_dir_name)
{
  if (cwd)
    {
      if (restore_cwd (cwd, "..", curr_dir_name))
	exit (1);
      free_cwd (cwd);
    }
  else if (chdir ("..") < 0)
    {
    //  error (1, errno, _("cannot change to `..' from directory %s"),
	  //   quote (curr_dir_name));
    }
}

/* xgetcwd.c -- return current directory with unlimited length
   Copyright (C) 1992, 1996, 2000 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* Written by David MacKenzie <djm@gnu.ai.mit.edu>.  */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#ifndef errno
extern int errno;
#endif

#include <sys/types.h>
#include "pathmax.h"

#if HAVE_GETCWD
char *getcwd ();
#else
char *getwd ();
# define getcwd(Buf, Max) getwd (Buf)
#endif

/* Return the current directory, newly allocated, arbitrarily long.
   Return NULL and set errno on error. */

char *
xgetcwd ()
{
  char *ret;
  unsigned path_max;
  char buf[1024];

  errno = 0;
  ret = getcwd (buf, sizeof (buf));
  if (ret != NULL)
    return xstrdup (buf);
  if (errno != ERANGE)
    return NULL;

  path_max = (unsigned) PATH_MAX;
  path_max += 2;		/* The getcwd docs say to do this. */

  for (;;)
    {
      char *cwd = (char *) xmalloc (path_max);

      errno = 0;
      ret = getcwd (cwd, path_max);
      if (ret != NULL)
	return ret;
      if (errno != ERANGE)
	{
	  int save_errno = errno;
	  free (cwd);
	  errno = save_errno;
	  return NULL;
	}

      free (cwd);

      path_max += path_max / 16;
      path_max += 32;
    }
}

void
free_cwd (struct saved_cwd *cwd)
{
  if (cwd->desc >= 0)
    close (cwd->desc);
  if (cwd->name)
    free (cwd->name);
}

void (*xalloc_fail_func) PARAMS ((void)) = 0;

void
xalloc_die (void)
{
  if (xalloc_fail_func)
    (*xalloc_fail_func) ();
  //error (xalloc_exit_failure, 0, "%s", _(xalloc_msg_memory_exhausted));
  /* The `noreturn' cannot be given to error, since it may return if
     its first argument is 0.  To help compilers understand the
     xalloc_die does terminate, call exit. */
  exit (EXIT_FAILURE);
}



/* Return a freshly allocated string containing the filenames
   in directory DIR, separated by '\0' characters;
   the end is marked by two '\0' characters in a row.
   NAME_SIZE is the number of bytes to initially allocate
   for the string; it will be enlarged as needed.
   Use NAME_SIZE == -1 if you do not know the size.
   Return NULL (setting errno) if DIR cannot be opened, read, or closed.  */

#ifndef NAME_SIZE_DEFAULT
# define NAME_SIZE_DEFAULT 512
#endif

char *
savedir (const char *dir, off_t name_size)
{
  DIR *dirp;
  struct dirent *dp;
  char *name_space;
  size_t allocated = name_size; /* Overflow is checked indirectly below.  */
  size_t used = 0;
  int save_errno;

  dirp = opendir (dir);
  if (dirp == NULL)
    return NULL;

  /* Use the default if the size is not known.  Be sure "allocated"
     is at least `1' so there's room for the final NUL byte.
     Do not simply test name_size <= 0, because the initialization
     of "allocated" might have overflowed.  */
  if (name_size < 0 || allocated == 0)
    allocated = NAME_SIZE_DEFAULT;

  name_space = static_cast<char *>(xmalloc (allocated));

  errno = 0;
  while ((dp = readdir (dirp)) != NULL)
    {
      /* Skip "", ".", and "..".  "" is returned by at least one buggy
         implementation: Solaris 2.4 readdir on NFS filesystems.  */
      char const *entry = dp->d_name;
      if (entry[entry[0] != '.' ? 0 : entry[1] != '.' ? 1 : 2] != '\0')
	{
	  size_t entry_size = strlen (entry) + 1;
	  if (used + entry_size < used)
	    xalloc_die ();
	  if (allocated <= used + entry_size)
	    {
	      do
		{
		  if (2 * allocated < allocated)
		    xalloc_die ();
		  allocated *= 2;
		}
	      while (allocated <= used + entry_size);

	      name_space = static_cast<char *>(xrealloc (name_space, allocated));
	    }
	  memcpy (name_space + used, entry, entry_size);
	  used += entry_size;
	}
    }
  name_space[used] = '\0';
  save_errno = errno;
  if (CLOSEDIR (dirp) != 0)
    save_errno = errno;
  if (save_errno != 0)
    {
      free (name_space);
      errno = save_errno;
      return NULL;
    }
  return name_space;
}

/* Change to recorded location, CWD, in directory hierarchy.
   If "saved working directory", NULL))
   */

int
restore_cwd (const struct saved_cwd *cwd, const char *dest, const char *from)
{
  int fail = 0;
  if (cwd->desc >= 0)
    {
      if (fchdir (cwd->desc))
	{
	  /*error (0, errno, "cannot return to %s%s%s",
		 (dest ? dest : "saved working directory"),
		 (from ? " from " : ""),
		 (from ? from : ""));*/
	  fail = 1;
	}
    }
  else if (chdir (cwd->name) < 0)
    {
//      error (0, errno, "%s", cwd->name);
      fail = 1;
    }
  return fail;
}

/* Record the location of the current working directory in CWD so that
   the program may change to other directories and later use restore_cwd
   to return to the recorded location.  This function may allocate
   space using malloc (via xgetcwd) or leave a file descriptor open;
   use free_cwd to perform the necessary free or close.  Upon failure,
   no memory is allocated, any locally opened file descriptors are
   closed;  return non-zero -- in that case, free_cwd need not be
   called, but doing so is ok.  Otherwise, return zero.  */

int
save_cwd (struct saved_cwd *cwd)
{
  static int have_working_fchdir = 1;

  cwd->desc = -1;
  cwd->name = NULL;

  if (have_working_fchdir)
    {
#if HAVE_FCHDIR
      cwd->desc = open (".", O_RDONLY | O_DIRECTORY);
      if (cwd->desc < 0)
	{
	  //error (0, errno, "cannot open current directory");
	  return 1;
	}

# if __sun__ || sun
      /* On SunOS 4, fchdir returns EINVAL if accounting is enabled,
	 so we have to fall back to chdir.  */
      if (fchdir (cwd->desc))
	{
	  if (errno == EINVAL)
	    {
	      close (cwd->desc);
	      cwd->desc = -1;
	      have_working_fchdir = 0;
	    }
	  else
	    {
	      error (0, errno, "current directory");
	      close (cwd->desc);
	      cwd->desc = -1;
	      return 1;
	    }
	}
# endif /* __sun__ || sun */
#else
# define fchdir(x) (abort (), 0)
      have_working_fchdir = 0;
#endif
    }

  if (!have_working_fchdir)
    {
      cwd->name = xgetcwd ();
      if (cwd->name == NULL)
	{
//	  error (0, errno, "cannot get current directory");
	  return 1;
	}
    }
  return 0;
}
/* Bits set in the FLAGS argument to `fnmatch'.  */
#define	FNM_PATHNAME	(1 << 0) /* No wildcard can ever match `/'.  */
#define	FNM_NOESCAPE	(1 << 1) /* Backslashes don't quote special chars.  */
#define	FNM_PERIOD	(1 << 2) /* Leading `.' is matched only explicitly.  */

#if !defined (_POSIX_C_SOURCE) || _POSIX_C_SOURCE < 2 || defined (_GNU_SOURCE)
#define	FNM_FILE_NAME	FNM_PATHNAME /* Preferred GNU name.  */
#define	FNM_LEADING_DIR	(1 << 3) /* Ignore `/...' after a match.  */
#define	FNM_CASEFOLD	(1 << 4) /* Compare without regard to case.  */
#endif

/* Value returned by `fnmatch' if STRING does not match PATTERN.  */
#define	FNM_NOMATCH	1



/* Match STRING against the filename pattern PATTERN, returning zero if
   it matches, nonzero if not.  */
int
fnmatch (const char *pattern, const char *string, int flags)
{
  register const char *p = pattern, *n = string;
  register char c;

/* Note that this evaluates C many times.  */
#define FOLD(c) ((flags & FNM_CASEFOLD) && ISUPPER ((unsigned char) (c)) \
                 ? tolower ((unsigned char) (c)) \
                 : (c))

  while ((c = *p++) != '\0')
    {
      c = FOLD (c);

      switch (c)
	{
	case '?':
	  if (*n == '\0')
	    return FNM_NOMATCH;
	  else if ((flags & FNM_FILE_NAME) && *n == '/')
	    return FNM_NOMATCH;
	  else if ((flags & FNM_PERIOD) && *n == '.' &&
		   (n == string || ((flags & FNM_FILE_NAME) && n[-1] == '/')))
	    return FNM_NOMATCH;
	  break;

	case '\\':
	  if (!(flags & FNM_NOESCAPE))
	    {
	      c = *p++;
	      if (c == '\0')
		/* Trailing \ loses.  */
		return FNM_NOMATCH;
	      c = FOLD (c);
	    }
	  if (FOLD (*n) != c)
	    return FNM_NOMATCH;
	  break;

	case '*':
	  if ((flags & FNM_PERIOD) && *n == '.' &&
	      (n == string || ((flags & FNM_FILE_NAME) && n[-1] == '/')))
	    return FNM_NOMATCH;

	  for (c = *p++; c == '?' || c == '*'; c = *p++)
	    {
	      if (c == '?')
		{
		  /* A ? needs to match one character.  */
		  if (*n == '\0' || (*n == '/' && (flags & FNM_FILE_NAME)))
		    /* There isn't another character; no match.  */
		    return FNM_NOMATCH;
		  else
		    /* One character of the string is consumed in matching
		       this ? wildcard, so *??? won't match if there are
		       less than three characters.  */
		    ++n;
		}
	    }

	  if (c == '\0')
	    {
	      if ((flags & (FNM_FILE_NAME | FNM_LEADING_DIR)) == FNM_FILE_NAME)
		for (; *n != '\0'; n++)
		  if (*n == '/')
		    return FNM_NOMATCH;
	      return 0;
	    }

	  {
	    char c1 = (!(flags & FNM_NOESCAPE) && c == '\\') ? *p : c;
	    c1 = FOLD (c1);
	    for (--p; *n != '\0'; ++n)
	      if ((c == '[' || FOLD (*n) == c1) &&
		  fnmatch (p, n, flags & ~FNM_PERIOD) == 0)
		return 0;
	      else if (*n == '/' && (flags & FNM_FILE_NAME))
		break;
	    return FNM_NOMATCH;
	  }

	case '[':
	  {
	    /* Nonzero if the sense of the character class is inverted.  */
	    register int not;

	    if (*n == '\0')
	      return FNM_NOMATCH;

	    if ((flags & FNM_PERIOD) && *n == '.' &&
		(n == string || ((flags & FNM_FILE_NAME) && n[-1] == '/')))
	      return FNM_NOMATCH;

	    not = (*p == '!' || *p == '^');
	    if (not)
	      ++p;

	    c = *p++;
	    for (;;)
	      {
		register char cstart = c, cend = c;

		if (!(flags & FNM_NOESCAPE) && c == '\\')
		  {
		    if (*p == '\0')
		      return FNM_NOMATCH;
		    cstart = cend = *p++;
		  }

		cstart = cend = FOLD (cstart);

		if (c == '\0')
		  /* [ (unterminated) loses.  */
		  return FNM_NOMATCH;

		c = *p++;
		c = FOLD (c);

		if ((flags & FNM_FILE_NAME) && c == '/')
		  /* [/] can never match.  */
		  return FNM_NOMATCH;

		if (c == '-' && *p != ']')
		  {
		    cend = *p++;
		    if (!(flags & FNM_NOESCAPE) && cend == '\\')
		      cend = *p++;
		    if (cend == '\0')
		      return FNM_NOMATCH;
		    cend = FOLD (cend);

		    c = *p++;
		  }

		if (FOLD (*n) >= cstart && FOLD (*n) <= cend)
		  goto matched;

		if (c == ']')
		  break;
	      }
	    if (!not)
	      return FNM_NOMATCH;
	    break;

	  matched:;
	    /* Skip the rest of the [...] that already matched.  */
	    while (c != ']')
	      {
		if (c == '\0')
		  /* [... (unterminated) loses.  */
		  return FNM_NOMATCH;

		c = *p++;
		if (!(flags & FNM_NOESCAPE) && c == '\\')
		  {
		    if (*p == '\0')
		      return FNM_NOMATCH;
		    /* XXX 1003.2d11 is unclear if this is right.  */
		    ++p;
		  }
	      }
	    if (not)
	      return FNM_NOMATCH;
	  }
	  break;

	default:
	  if (c != FOLD (*n))
	    return FNM_NOMATCH;
	}

      ++n;
    }

  if (*n == '\0')
    return 0;

  if ((flags & FNM_LEADING_DIR) && *n == '/')
    /* The FNM_LEADING_DIR flag says that "foo*" matches "foobar/frobozz".  */
    return 0;

  return FNM_NOMATCH;

#undef FOLD
}


struct exclude_st
  {
    char const **exclude;
    int exclude_alloc;
    int exclude_count;
  };

struct exclude_st *
new_exclude (void)
{
  struct exclude_st *ex = (struct exclude_st *) xmalloc (sizeof (struct exclude_st));
  ex->exclude_count = 0;
  ex->exclude_alloc = 64;
  ex->exclude = (char const **) xmalloc (ex->exclude_alloc * sizeof (char *));
  return ex;
}

int
excluded_filename (struct exclude_st const *ex, char const *f, int options)
{
  char const * const *exclude = ex->exclude;
  int exclude_count = ex->exclude_count;
  int i;

  for (i = 0;  i < exclude_count;  i++)
    if (fnmatch (exclude[i], f, options) == 0)
      return 1;

  return 0;
}

/* Allocate space for the hash structures, and set the global
   variable `htab' to point to it.  The initial hash module is specified in
   MODULUS, and the number of entries are specified in ENTRY_TAB_SIZE.  (The
   hash structure will be rebuilt when ENTRY_TAB_SIZE entries have been
   inserted, and MODULUS and ENTRY_TAB_SIZE in the global `htab' will be
   doubled.)  */

void
DirTree::hash_init (unsigned int modulus, unsigned int entry_tab_size)
{
  struct htab *htab_r;

  htab_r = (struct htab *)
    xmalloc (sizeof (struct htab) + sizeof (struct entry *) * modulus);

  htab_r->entry_tab = (struct entry *)
    xmalloc (sizeof (struct entry) * entry_tab_size);

  htab_r->modulus = modulus;
  htab_r->entry_tab_size = entry_tab_size;
  htab = htab_r;

  hash_reset ();
}

void DirTree::hash_reset (void)
{
  int i;
  struct entry **p;

  htab->first_free_entry = 0;

  p = htab->hash;
  for (i = htab->modulus; i > 0; i--)
    *p++ = NULL;
}

/* Insert INO and DEV in the hash structure HTAB, if not
   already present.  Return zero if inserted and nonzero if it
   already existed.  */

int DirTree::hash_insert2 (struct htab *ht, ino_t ino, dev_t dev)
{
  struct entry **hp, *ep2, *ep;
  hp = &ht->hash[ino % ht->modulus];
  ep2 = *hp;

  /* Collision?  */

  if (ep2 != NULL)
    {
      ep = ep2;

      /* Search for an entry with the same data.  */

      do
	{
	  if (ep->ino == ino && ep->dev == dev)
	    return 1;		/* Found an entry with the same data.  */
	  ep = ep->coll_link;
	}
      while (ep != NULL);

      /* Did not find it.  */

    }

  ep = *hp = &ht->entry_tab[ht->first_free_entry++];
  ep->ino = ino;
  ep->dev = dev;
  ep->coll_link = ep2;		/* `ep2' is NULL if no collision.  */

  return 0;
}


/* Insert an item (inode INO and device DEV) in the hash
   structure in the global variable `htab', if an entry with the same data
   was not found already.  Return zero if the item was inserted and nonzero
   if it wasn't.  */

int DirTree::hash_insert (ino_t ino, dev_t dev)
{
  struct htab *htab_r = htab;	/* Initially a copy of the global `htab'.  */

  if (htab_r->first_free_entry >= htab_r->entry_tab_size)
    {
      int i;
      struct entry *ep;
      unsigned modulus;
      unsigned entry_tab_size;

      /* Increase the number of hash entries, and re-hash the data.
	 The method of shrimping and increasing is made to compactify
	 the heap.  If twice as much data would be allocated
	 straightforwardly, we would never re-use a byte of memory.  */

      /* Let `htab' shrimp.  Keep only the header, not the pointer vector.  */

      htab_r = (struct htab *)
	xrealloc ((char *) htab_r, sizeof (struct htab));

      modulus = 2 * htab_r->modulus;
      entry_tab_size = 2 * htab_r->entry_tab_size;

      /* Increase the number of possible entries.  */

      htab_r->entry_tab = (struct entry *)
	xrealloc ((char *) htab_r->entry_tab,
		  sizeof (struct entry) * entry_tab_size);

      /* Increase the size of htab again.  */

      htab_r = (struct htab *)
	xrealloc ((char *) htab_r,
		 sizeof (struct htab) + sizeof (struct entry *) * modulus);

      htab_r->modulus = modulus;
      htab_r->entry_tab_size = entry_tab_size;
      htab = htab_r;

      i = htab_r->first_free_entry;

      /* Make the increased hash table empty.  The entries are still
	 available in htab->entry_tab.  */

      hash_reset ();

      /* Go through the entries and install them in the pointer vector
	 htab->hash.  The items are actually inserted in htab->entry_tab at
	 the position where they already are.  The htab->coll_link need
	 however be updated.  Could be made a little more efficient.  */

      for (ep = htab_r->entry_tab; i > 0; i--)
	{
	  hash_insert2 (htab_r, ep->ino, ep->dev);
	  ep++;
	}
    }

  return hash_insert2 (htab_r, ino, dev);
}



/* determine the size 
   of file or directory ENT. Return the size of ENT in units of 512-byte
   blocks.  TOP is one for external calls, zero for recursive calls.
   LAST_DEV is the device that the parent directory of ENT is on.
   DEPTH is the number of levels (in hierarchy) down from a command
   line argument.  Don't print if DEPTH > max_depth.
   An important invariant is that when this function returns, the current
   working directory is the same as when it was called.  */

/* Now stores the information in en_tree nodes if appropriate.  If target is
   NULL, prints results immediately.  If target is not NULL, stores the results
   in target (and children) */
uintmax_t DirTree::count_entry (const char *ent, int top, dev_t last_dev,
	int depth, en_tree *target)
{
  uintmax_t size;
  struct stat stat_buf;
  en_tree *unused_child=NULL;

  if (((top && opt_dereference_arguments)
       ? stat (ent, &stat_buf)
       : (*xstat) (ent, &stat_buf)) < 0)
    {
      //error (0, errno, "%s", path->text);
      exit_status = 1;
      return 0;
    }

  if (!opt_count_all
      && stat_buf.st_nlink > 1
      && hash_insert (stat_buf.st_ino, stat_buf.st_dev))
    return 0;			/* Have counted this already.  */

  size = ST_NBLOCKS (stat_buf);
  tot_size += size;

  if (S_ISDIR (stat_buf.st_mode))
    {
      unsigned pathlen;
      dev_t dir_dev;
      char *name_space;
      char *namep;
      struct saved_cwd *cwd;
      struct saved_cwd cwd_buf;
      struct stat e_buf;

      dir_dev = stat_buf.st_dev;

      /* Return `0' here, not SIZE, since the SIZE bytes
	 would reside in the new filesystem.  */
      if (opt_one_file_system && !top && last_dev != dir_dev)
	return 0;		/* Don't enter a new file system.  */

#ifndef S_ISLNK
# define S_ISLNK(s) 0
#endif
      /* If we're traversing more than one level, or if we're
	 dereferencing symlinks and we're about to chdir through a
	 symlink, remember the current directory so we can return to
	 it later.  In other cases, chdir ("..") works fine.
	 Treat `.' and `..' like multi-level paths, since `chdir ("..")'
	 wont't restore the current working directory after a `chdir'
	 to one of those.  */
      if (strchr (ent, '/')
	  || DOT_OR_DOTDOT (ent)
	  || (xstat == stat
	      && lstat (ent, &e_buf) == 0
	      && S_ISLNK (e_buf.st_mode)))
	{
	  if (save_cwd (&cwd_buf))
	    exit (1);
	  cwd = &cwd_buf;
	}
      else
	cwd = NULL;

      if (chdir (ent) < 0)
	{
	  //error (0, errno, _("cannot change to directory %s"),
		 //path->text);
	  if (cwd)
	    free_cwd (cwd);
	  exit_status = 1;
	  /* Do return SIZE, here, since even though we can't chdir into ENT,
	     we *can* count the blocks used by its directory entry.  */
	  return propogate_up(opt_separate_dirs ? 0 : size, target);
	}

      name_space = savedir (".", stat_buf.st_size);
      if (name_space == NULL)
	{
	  //error (0, errno, "%s", path->text);
	  pop_dir (cwd, path->text);
	  exit_status = 1;
	  /* Do count the SIZE bytes.  */
	  return propogate_up(opt_separate_dirs ? 0 : size, target);
	}

      /* Remember the current path.  */

      str_concatc (path, "/");
      pathlen = path->length;

      for (namep = name_space; *namep; namep += strlen (namep) + 1)
	{
	  if (!excluded_filename (exclude, namep, 0))
	    {
	      en_tree *child=unused_child;
	      if (child==NULL && target !=NULL)
	      {
		child=en_tree_new();
		en_tree_add_child(target, child);
	      }
	      str_concatc (path, namep);
	      size += count_entry (namep, 0, dir_dev, depth + 1, child);
	      if (child!=NULL && child->name==NULL) 
		unused_child=child;
	      else unused_child=NULL;
	      str_trunc (path, pathlen);
	    }
	}

      en_tree_remove(unused_child);
      free (name_space);
      pop_dir (cwd, path->text);

      str_trunc (path, pathlen - 1); /* Remove the "/" we added.  */
      if (depth <= max_depth || top)
      {
	if (target==NULL) /*print_size (size, path->length > 0 ? path->text : "/")*/;
	else en_tree_set(target, size, path->length > 0 ? path->text : "/");
      }
      return propogate_up(opt_separate_dirs ? 0 : size, target);
    }
  else if ((opt_all && depth <= max_depth) || top)
    {
      /* FIXME: make this an option.  */
      int print_only_dir_size = 0;
      if (!print_only_dir_size)
      {
	if (target==NULL) /*print_size (size, path->length > 0 ? path->text : "/")*/;
	else en_tree_set(target, size, path->length > 0 ? path->text : "/");
      }

    }

  return propogate_up(size, target);
}

uintmax_t DirTree::propogate_up(uintmax_t size, en_tree *target)
{
	if (target!=NULL && up_immediate) 
	{
		en_tree_incr_up(target, size);
		return 0;
	}
	else return size;
}

DirTree::DirTree()
{
	init();
}

void DirTree::init()
{
	max_depth=INT_MAX;
	exclude=new_exclude();
	path=NULL;
	str_init(&path, INITIAL_PATH_SIZE);
	xstat=lstat;
	htab=NULL;
	tot_size=0;
	hash_init(INITIAL_HASH_MODULE, INITIAL_ENTRY_TAB_SIZE);

	opt_count_all=false;
	opt_dereference_arguments=false;
	opt_one_file_system=false;
	opt_separate_dirs=false;
	opt_all =false;
	up_immediate=false;
}

en_tree *DirTree::get_en_tree(const char *target)
{
	en_tree *my_tree;
	my_tree=en_tree_new();
	count_entry(target, 1, 0, 0, my_tree);
	return my_tree;
}

HelloWorld::HelloWorld():
	Gtk::Window(GTK_WINDOW_TOPLEVEL),
	m_button("Exit"),
	child_button("Child"),
	next_button("Next"),
	prev_button("Prev"),
	parent_button("Parent"),
	tview(300, 300),
	curdir("./")
{
	destroy.connect(slot(destroy_handler));
	set_border_width(10);
	m_button.clicked.connect(destroy.slot());
	child_button.clicked.connect(slot(this, &HelloWorld::nav_child));
	next_button.clicked.connect(slot(this, &HelloWorld::nav_next));
	parent_button.clicked.connect(slot(this, &HelloWorld::nav_parent));
	prev_button.clicked.connect(slot(this, &HelloWorld::nav_prev));
	

	add(packer);
	packer.pack_start(curdir);
	packer.pack_start(m_button);
	
	packer.pack_start(parent_button);
	packer.pack_start(prev_button);
	packer.pack_start(next_button);
	packer.pack_start(child_button);
	packer.pack_start(tview);
	curdir.show();
	m_button.show();
	parent_button.show();
	child_button.show();
	next_button.show();
	prev_button.show();
	tview.show();
	packer.show();
	show();
}

void HelloWorld::hello()
{
	cout << "Hello World!" << endl;
}

int HelloWorld::delete_event_impl(GdkEventAny *)
{
	cout << "A delete event ocurred" << endl;
	return false;
}

int main (int argc, char *argv[])
{
  // all GTK applications must have a gtk_main(). Control ends here
  // and waits for an event to occur (like a key press or mouse event).
  Gtk::Main kit(argc, argv);

  HelloWorld helloworld;

  kit.run();
  return 0;
}


