#if HAVE_DIRENT_H
# include <dirent.h>
#else
# define dirent direct
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif


#define DOT_OR_DOTDOT(Basename) \
  (Basename[0] == '.' && (Basename[1] == '\0' \
			  || (Basename[1] == '.' && Basename[2] == '\0')))

#define xstrdup strdup
#define xrealloc realloc
/* Extract or fake data from a `struct stat'.
   ST_BLKSIZE: Preferred I/O blocksize for the file, in bytes.
   ST_NBLOCKS: Number of blocks in the file, including indirect blocks.
   ST_NBLOCKSIZE: Size of blocks used when calculating ST_NBLOCKS.  */
#ifndef HAVE_STRUCT_STAT_ST_BLOCKS
# define ST_BLKSIZE(statbuf) DEV_BSIZE
# if defined(_POSIX_SOURCE) || !defined(BSIZE) /* fileblocks.c uses BSIZE.  */
#  define ST_NBLOCKS(statbuf) \
  (S_ISREG ((statbuf).st_mode) \
   || S_ISDIR ((statbuf).st_mode) \
   ? (statbuf).st_size / ST_NBLOCKSIZE + ((statbuf).st_size % ST_NBLOCKSIZE != 0) : 0)
# else /* !_POSIX_SOURCE && BSIZE */
#  define ST_NBLOCKS(statbuf) \
  (S_ISREG ((statbuf).st_mode) \
   || S_ISDIR ((statbuf).st_mode) \
   ? st_blocks ((statbuf).st_size) : 0)
# endif /* !_POSIX_SOURCE && BSIZE */
#else /* HAVE_STRUCT_STAT_ST_BLOCKS */
/* Some systems, like Sequents, return st_blksize of 0 on pipes. */
# define ST_BLKSIZE(statbuf) ((statbuf).st_blksize > 0 \
			       ? (statbuf).st_blksize : DEV_BSIZE)
# if defined(hpux) || defined(__hpux__) || defined(__hpux)
/* HP-UX counts st_blocks in 1024-byte units.
   This loses when mixing HP-UX and BSD filesystems with NFS.  */
#  define ST_NBLOCKSIZE 1024
# else /* !hpux */
#  if defined(_AIX) && defined(_I386)
/* AIX PS/2 counts st_blocks in 4K units.  */
#   define ST_NBLOCKSIZE (4 * 1024)
#  else /* not AIX PS/2 */
#   if defined(_CRAY)
#    define ST_NBLOCKS(statbuf) \
  (S_ISREG ((statbuf).st_mode) \
   || S_ISDIR ((statbuf).st_mode) \
   ? (statbuf).st_blocks * ST_BLKSIZE(statbuf)/ST_NBLOCKSIZE : 0)
#   endif /* _CRAY */
#  endif /* not AIX PS/2 */
# endif /* !hpux */
#endif /* HAVE_STRUCT_STAT_ST_BLOCKS */

#ifndef ST_NBLOCKS
# define ST_NBLOCKS(statbuf) \
  (S_ISREG ((statbuf).st_mode) \
   || S_ISDIR ((statbuf).st_mode) \
   ? (statbuf).st_blocks : 0)
#endif

#ifndef ST_NBLOCKSIZE
# define ST_NBLOCKSIZE 512
#endif

#ifndef O_DIRECTORY
# define O_DIRECTORY 0
#endif

#if CLOSEDIR_VOID
/* Fake a return value. */
# define CLOSEDIR(d) (closedir (d), 0)
#else
# define CLOSEDIR(d) closedir (d)
#endif

#if STDC_HEADERS || (!defined (isascii) && !HAVE_ISASCII)
# define IN_CTYPE_DOMAIN(c) 1
#else
# define IN_CTYPE_DOMAIN(c) isascii(c)
#endif

#define ISUPPER(c) (IN_CTYPE_DOMAIN (c) && isupper (c))

using std::cout;
using std::endl;

using SigC::slot;

template<typename T>
class Watcher
{
	public:
	virtual void modified(T const &modobj)=0;
};


class Rectangle
{
	public:
	float left, top, width, height;
	Rectangle(float left, float top, float width, float height);
	bool contains(float x, float y);
};

class TextRect: public Rectangle
{
	private:
	en_tree *my_en_tree;
	public:
	TextRect(float left, float top, float width, float height):
		Rectangle(left, top, width, height)
	{;}
	TextRect(TextRect const &arg);
	void setText(en_tree *arg_en_tree);
	en_tree *get_en_tree()
	{return my_en_tree;}
	void setColor(Gdk_Color &curcolor);
};

struct String
{
  unsigned alloc;		/* Size of allocation for the text.  */
  unsigned length;		/* Length of the text currently.  */
  char *text;			/* Pointer to the text.  */
};
typedef struct String String;


class DirTree
{

	struct exclude_st *exclude;
	String *path;
	int (*xstat) (const char *, struct stat *buf);
	int exit_status;
	uintmax_t tot_size;
	struct htab *htab;

/* Allocate space for the hash structures, and set the global
   variable `htab' to point to it.  The initial hash module is specified in
   MODULUS, and the number of entries are specified in ENTRY_TAB_SIZE.  (The
   hash structure will be rebuilt when ENTRY_TAB_SIZE entries have been
   inserted, and MODULUS and ENTRY_TAB_SIZE in the global `htab' will be
   doubled.)  */

	void hash_init (unsigned int modulus, unsigned int entry_tab_size);
	int hash_insert2 (struct htab *ht, ino_t ino, dev_t dev);
	int hash_insert (ino_t ino, dev_t dev);
	void hash_reset (void);

	uintmax_t count_entry (const char *ent, int top, dev_t last_dev, int depth, 
	en_tree *target);
	uintmax_t propogate_up(uintmax_t size, en_tree *target);

	void init();
	Watcher<DirTree> *my_watch;
	public:
	bool opt_count_all;
	bool opt_dereference_arguments;
	bool opt_one_file_system;
	bool opt_separate_dirs;
// Print all entries, files and directories
	bool opt_all;
	bool up_immediate;

	int max_depth;

	DirTree();
	void calc_en_tree(const char *target, en_tree *my_tree);
	en_tree *get_en_tree(const char *);
	void set_watcher(Watcher<DirTree> &my_watch);

};


class TreeView: public Gtk::DrawingArea, public Watcher<DirTree>
{
	en_tree *my_tree;
	en_tree *cur_tree;
	TextRect *over_rect;
	vector<TextRect> rectangles;
	bool firsttime;
	uintmax_t oldsize;
	Watcher<TreeView> *my_watch;

	DirTree treeGen;
	public:
	TreeView(gint x_size = 0, gint y_size = 0);
	en_tree const *get_en_tree();
	void set_en_tree(en_tree *newcur);
	en_tree const *get_top_en_tree();
	void modified(DirTree const &moddir);
	void me_modified()
	{
		if (my_watch!=NULL) my_watch->modified(*this);
	}
	gint expose_event_impl(GdkEventExpose*);
	gint motion_notify_event_impl(GdkEventMotion *event);
   gint button_press_event_impl(GdkEventButton *event);
	gint leave_notify_event_impl(GdkEventCrossing *event);

	void en_tree_rect(en_tree *curent, bool vertical, gint x, gint y, 
	gint width, gint height);
	void draw_rect();
	void seek_over(int x, int y);
	void test_tree();
	void set_watcher(Watcher<TreeView> &my_watch);
};

/* Initial number of entries in each hash table entry's table of inodes.  */
#define INITIAL_HASH_MODULE 100

/* Initial number of entries in the inode hash table.  */
#define INITIAL_ENTRY_TAB_SIZE 70

/* Initial size to allocate for `path'.  */
#define INITIAL_PATH_SIZE 100

class MainWindow: public Gtk::Window, public Watcher<TreeView>
{
	private:
	Gtk::Button child_button;
	Gtk::Button next_button;
	Gtk::Button prev_button;
	Gtk::Button parent_button;
	Gtk::Label curdir;
	Gtk::VBox packer;
	Gtk::VBox buttons;
	Gtk::HBox hbox;
	Gtk::HBox buttonslabel;
	TreeView tview;
	void nav(en_tree &en_new)
	{
		//setInfo(&en_new, tview.get_top_en_tree());
		tview.set_en_tree(&en_new);
	}
	void nav_child()
	{
		en_tree *child=tview.get_en_tree()->child;
		if (child!=NULL) nav(*child);
	}
	void nav_next()
	{
		en_tree *next=tview.get_en_tree()->next;
		if (next!=NULL) nav(*next);
	}
	void nav_prev()
	{
		en_tree *prev=tview.get_en_tree()->prev;
		if (prev!=NULL) nav(*prev);
	}
	void nav_parent()
	{
		en_tree *parent=tview.get_en_tree()->parent;
		if (parent!=NULL) nav(*parent);
	}
	int setInfo(en_tree const *cur, en_tree const *top);

	public:
	MainWindow();
	void hello();
	virtual int delete_event_impl(GdkEventAny *event);
	void modified(const TreeView &);
};

struct entry
{
  ino_t ino;
  dev_t dev;
  struct entry *coll_link;
};

struct htab
{
  unsigned modulus;		/* Size of the `hash' pointer vector.  */
  struct entry *entry_tab;	/* Pointer to dynamically growing vector.  */
  unsigned entry_tab_size;	/* Size of current `entry_tab' allocation.  */
  unsigned first_free_entry;	/* Index in `entry_tab'.  */
  struct entry *hash[1];	/* Vector of pointers in `entry_tab'.  */
};

